/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __SRC_NM_SSTP_PPPD_SERVICE_DBUS_H__
#define __SRC_NM_SSTP_PPPD_SERVICE_DBUS_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.freedesktop.NetworkManager.sstp.ppp */

#define NMDBUS_TYPE_SSTP_PPP (nmdbus_sstp_ppp_get_type ())
#define NMDBUS_SSTP_PPP(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), NMDBUS_TYPE_SSTP_PPP, NMDBusSstpPpp))
#define NMDBUS_IS_SSTP_PPP(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), NMDBUS_TYPE_SSTP_PPP))
#define NMDBUS_SSTP_PPP_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), NMDBUS_TYPE_SSTP_PPP, NMDBusSstpPppIface))

struct _NMDBusSstpPpp;
typedef struct _NMDBusSstpPpp NMDBusSstpPpp;
typedef struct _NMDBusSstpPppIface NMDBusSstpPppIface;

struct _NMDBusSstpPppIface
{
  GTypeInterface parent_iface;

  gboolean (*handle_need_secrets) (
    NMDBusSstpPpp *object,
    GDBusMethodInvocation *invocation);

  gboolean (*handle_set_config) (
    NMDBusSstpPpp *object,
    GDBusMethodInvocation *invocation,
    GVariant *arg_config);

  gboolean (*handle_set_ip4_config) (
    NMDBusSstpPpp *object,
    GDBusMethodInvocation *invocation,
    GVariant *arg_config);

  gboolean (*handle_set_ip6_config) (
    NMDBusSstpPpp *object,
    GDBusMethodInvocation *invocation,
    GVariant *arg_config);

  gboolean (*handle_set_state) (
    NMDBusSstpPpp *object,
    GDBusMethodInvocation *invocation,
    guint arg_state);

};

GType nmdbus_sstp_ppp_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *nmdbus_sstp_ppp_interface_info (void);
guint nmdbus_sstp_ppp_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void nmdbus_sstp_ppp_complete_need_secrets (
    NMDBusSstpPpp *object,
    GDBusMethodInvocation *invocation,
    const gchar *username,
    const gchar *password);

void nmdbus_sstp_ppp_complete_set_config (
    NMDBusSstpPpp *object,
    GDBusMethodInvocation *invocation);

void nmdbus_sstp_ppp_complete_set_ip4_config (
    NMDBusSstpPpp *object,
    GDBusMethodInvocation *invocation);

void nmdbus_sstp_ppp_complete_set_ip6_config (
    NMDBusSstpPpp *object,
    GDBusMethodInvocation *invocation);

void nmdbus_sstp_ppp_complete_set_state (
    NMDBusSstpPpp *object,
    GDBusMethodInvocation *invocation);



/* D-Bus method calls: */
void nmdbus_sstp_ppp_call_need_secrets (
    NMDBusSstpPpp *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean nmdbus_sstp_ppp_call_need_secrets_finish (
    NMDBusSstpPpp *proxy,
    gchar **out_username,
    gchar **out_password,
    GAsyncResult *res,
    GError **error);

gboolean nmdbus_sstp_ppp_call_need_secrets_sync (
    NMDBusSstpPpp *proxy,
    gchar **out_username,
    gchar **out_password,
    GCancellable *cancellable,
    GError **error);

void nmdbus_sstp_ppp_call_set_config (
    NMDBusSstpPpp *proxy,
    GVariant *arg_config,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean nmdbus_sstp_ppp_call_set_config_finish (
    NMDBusSstpPpp *proxy,
    GAsyncResult *res,
    GError **error);

gboolean nmdbus_sstp_ppp_call_set_config_sync (
    NMDBusSstpPpp *proxy,
    GVariant *arg_config,
    GCancellable *cancellable,
    GError **error);

void nmdbus_sstp_ppp_call_set_ip4_config (
    NMDBusSstpPpp *proxy,
    GVariant *arg_config,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean nmdbus_sstp_ppp_call_set_ip4_config_finish (
    NMDBusSstpPpp *proxy,
    GAsyncResult *res,
    GError **error);

gboolean nmdbus_sstp_ppp_call_set_ip4_config_sync (
    NMDBusSstpPpp *proxy,
    GVariant *arg_config,
    GCancellable *cancellable,
    GError **error);

void nmdbus_sstp_ppp_call_set_ip6_config (
    NMDBusSstpPpp *proxy,
    GVariant *arg_config,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean nmdbus_sstp_ppp_call_set_ip6_config_finish (
    NMDBusSstpPpp *proxy,
    GAsyncResult *res,
    GError **error);

gboolean nmdbus_sstp_ppp_call_set_ip6_config_sync (
    NMDBusSstpPpp *proxy,
    GVariant *arg_config,
    GCancellable *cancellable,
    GError **error);

void nmdbus_sstp_ppp_call_set_state (
    NMDBusSstpPpp *proxy,
    guint arg_state,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean nmdbus_sstp_ppp_call_set_state_finish (
    NMDBusSstpPpp *proxy,
    GAsyncResult *res,
    GError **error);

gboolean nmdbus_sstp_ppp_call_set_state_sync (
    NMDBusSstpPpp *proxy,
    guint arg_state,
    GCancellable *cancellable,
    GError **error);



/* ---- */

#define NMDBUS_TYPE_SSTP_PPP_PROXY (nmdbus_sstp_ppp_proxy_get_type ())
#define NMDBUS_SSTP_PPP_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), NMDBUS_TYPE_SSTP_PPP_PROXY, NMDBusSstpPppProxy))
#define NMDBUS_SSTP_PPP_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), NMDBUS_TYPE_SSTP_PPP_PROXY, NMDBusSstpPppProxyClass))
#define NMDBUS_SSTP_PPP_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), NMDBUS_TYPE_SSTP_PPP_PROXY, NMDBusSstpPppProxyClass))
#define NMDBUS_IS_SSTP_PPP_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), NMDBUS_TYPE_SSTP_PPP_PROXY))
#define NMDBUS_IS_SSTP_PPP_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), NMDBUS_TYPE_SSTP_PPP_PROXY))

typedef struct _NMDBusSstpPppProxy NMDBusSstpPppProxy;
typedef struct _NMDBusSstpPppProxyClass NMDBusSstpPppProxyClass;
typedef struct _NMDBusSstpPppProxyPrivate NMDBusSstpPppProxyPrivate;

struct _NMDBusSstpPppProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  NMDBusSstpPppProxyPrivate *priv;
};

struct _NMDBusSstpPppProxyClass
{
  GDBusProxyClass parent_class;
};

GType nmdbus_sstp_ppp_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (NMDBusSstpPppProxy, g_object_unref)
#endif

void nmdbus_sstp_ppp_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
NMDBusSstpPpp *nmdbus_sstp_ppp_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
NMDBusSstpPpp *nmdbus_sstp_ppp_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void nmdbus_sstp_ppp_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
NMDBusSstpPpp *nmdbus_sstp_ppp_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
NMDBusSstpPpp *nmdbus_sstp_ppp_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define NMDBUS_TYPE_SSTP_PPP_SKELETON (nmdbus_sstp_ppp_skeleton_get_type ())
#define NMDBUS_SSTP_PPP_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), NMDBUS_TYPE_SSTP_PPP_SKELETON, NMDBusSstpPppSkeleton))
#define NMDBUS_SSTP_PPP_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), NMDBUS_TYPE_SSTP_PPP_SKELETON, NMDBusSstpPppSkeletonClass))
#define NMDBUS_SSTP_PPP_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), NMDBUS_TYPE_SSTP_PPP_SKELETON, NMDBusSstpPppSkeletonClass))
#define NMDBUS_IS_SSTP_PPP_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), NMDBUS_TYPE_SSTP_PPP_SKELETON))
#define NMDBUS_IS_SSTP_PPP_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), NMDBUS_TYPE_SSTP_PPP_SKELETON))

typedef struct _NMDBusSstpPppSkeleton NMDBusSstpPppSkeleton;
typedef struct _NMDBusSstpPppSkeletonClass NMDBusSstpPppSkeletonClass;
typedef struct _NMDBusSstpPppSkeletonPrivate NMDBusSstpPppSkeletonPrivate;

struct _NMDBusSstpPppSkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  NMDBusSstpPppSkeletonPrivate *priv;
};

struct _NMDBusSstpPppSkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType nmdbus_sstp_ppp_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (NMDBusSstpPppSkeleton, g_object_unref)
#endif

NMDBusSstpPpp *nmdbus_sstp_ppp_skeleton_new (void);


G_END_DECLS

#endif /* __SRC_NM_SSTP_PPPD_SERVICE_DBUS_H__ */
